use std::error::Error as StdError;

use anyhow::Result;
use async_trait::async_trait;
use bytes::Bytes;
use cookie::Cookie;
use hyper::{header, Body, Client, Method, Request, StatusCode};
use thiserror::Error;
use url::Url;

#[derive(Error, Debug)]
pub enum DownloadError {
    #[error("Http error")]
    HttpError(#[from] hyper::http::Error),
    #[error("Hyper error")]
    HyperError(#[from] hyper::Error),
}

/// Downloads documents
#[async_trait]
pub trait Download {
    type Error: StdError;

    /// Downloads a file from an url and returns the result as bytes
    ///
    /// If the file is not found, returns None
    async fn download(&self, file_link: &Url) -> Result<Option<Bytes>, Self::Error>;
}

/// Store several cookies
// TODO: add builder or new() funciton
#[derive(Debug, Clone)]
pub struct Downloader<'c> {
    pub cookies: Vec<Cookie<'c>>,
}

#[async_trait]
impl<'c> Download for Downloader<'c> {
    type Error = DownloadError;

    async fn download(&self, file_link: &Url) -> Result<Option<Bytes>, Self::Error> {
        log::debug!("downloading url {:?}", file_link);
        let https = hyper_rustls::HttpsConnector::with_native_roots();
        let client: Client<_, hyper::Body> = Client::builder().build(https);

        let mut req = Request::builder()
            .method(Method::GET)
            .uri(file_link.as_str());

        req = req.header(
            header::COOKIE,
            self.cookies
                .iter()
                .map(Cookie::to_string)
                .collect::<Vec<_>>()
                .join(";"),
        );
        log::debug!("headers : {:?}", req.headers_ref());

        let req = req.body(Body::empty())?;

        let resp = client.request(req).await?;
        log::debug!("Response status : {:?}", resp.status());
        let body = match resp.status() {
            StatusCode::OK => Some(hyper::body::to_bytes(resp).await?),
            StatusCode::NOT_FOUND => None,
            // TODO: enhance this by handling more error codes
            _ => None,
        };
        Ok(body)
    }
}
